/***************************************/
/*Program Name  : Direct Draw Utilities*/
/*File Name     : ddutils.cpp          */
/*Author        : Ivan Wright          */
/*Updated       : 26/03/2000           */
/*Version       : 2.3d                 */
/***************************************/
//---------------------------------------------------------------------------
#include "DDUtils.h"
//---------------------------------------------------------------------------
namespace TDDraw {
//---------------------------------------------------------------------------
void __fastcall DDBlitFast(IDirectDrawSurface *idds, LPDIRECTDRAWSURFACE lpdds,
 FPoint SPos, RECT Size)
{
  RECT R;
  HRESULT ddrval;
  _DDBLTFX BltFx;
  BltFx.dwDDFX = DDFXCAPS_OVERLAYMIRRORLEFTRIGHT;
  BltFx.dwSize = sizeof(BltFx);
  if (SPos.X < 0) {                            //Check if X < 164
    if (SPos.X + Size.right < 0) return;       //Check if X + width < 164
    Size.left = 0 - SPos.X;                    //Move Left position in
    SPos.X = 0;                                //Move X to 164
  }
  if (SPos.Y < 40) {                             //Check if Y < 40
    if (SPos.Y + Size.bottom < 40) return;       //Check if Y + height < 40
    Size.top = 40 - SPos.Y;                      //Move Top position in
    SPos.Y = 40;                                 //Move Y to 40
  }
  if (SPos.X > 1152) return;                      //Check if X goes off screen
  if (SPos.X + Size.right > 1152)                 //Move Right position back
   Size.right = 1152 - SPos.X;

  if (SPos.Y > 864) return;                      //Check if Y goes iff screen
  if (SPos.Y + Size.bottom > 864)                //Move bottom position up
   Size.bottom = 864 - SPos.Y;
  R.left = SPos.X;
  R.right = R.left + Size.right;
  R.top = SPos.Y;
  R.bottom = R.top + Size.bottom;
                                                 //Ready to draw the image
  ddrval = idds->BltFast(SPos.X, SPos.Y, lpdds, &Size, DDBLTFAST_SRCCOLORKEY);
  //ddrval = idds->Blt(&R, lpdds, &Size, DDBLT_KEYSRC, NULL);
  if (ddrval == DDERR_INVALIDRECT) return;
  DDError(ddrval);
}
//---------------------------------------------------------------------------
HRESULT DDCopyBitmap(IDirectDrawSurface *pdds, HBITMAP hbm, FPoint xy,
 FPoint dxy)
{
  HDC hdcImage;
  HDC hdc;
  BITMAP bm;
  DDSURFACEDESC ddsd;
  HRESULT hr;

  if (hbm == NULL || pdds == NULL) {
    ShowMessage("Can't copy bitmap");
    exit(EXIT_FAILURE);                          //exit the program
    return E_FAIL;
  }

  pdds->Restore();
  hdcImage = CreateCompatibleDC(NULL);
  if (!hdcImage) {
    ShowMessage("Create Compatible DC failed!");
    exit(EXIT_FAILURE);                          //exit the program
  }
  SelectObject(hdcImage, hbm);

  GetObject(hbm, sizeof(bm), &bm);               //get size of bitmap
  if (dxy.X == 0) dxy.X = bm.bmWidth;            //use the passed size, unless zero
  if (dxy.Y == 0) dxy.Y = bm.bmHeight;
  ddsd.dwSize = sizeof(ddsd);
  ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH;
  pdds->GetSurfaceDesc(&ddsd);

  if ((hr = pdds->GetDC(&hdc)) == DD_OK) {
    StretchBlt(hdc, 0, 0, ddsd.dwWidth, ddsd.dwHeight, hdcImage, xy.X, xy.Y,
     dxy.X, dxy.Y, SRCCOPY);
    pdds->ReleaseDC(hdc);
  }
  DeleteDC(hdcImage);
  return hr;
} //DDCopyBitmap
//---------------------------------------------------------------------------
IDirectDrawSurface *DDCreateSurface(IDirectDraw *pdd, DWORD Height, DWORD Width)
{
  DDSURFACEDESC ddsd;
  IDirectDrawSurface *pdds;

  ZeroMemory(&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof(ddsd);                    //Setup Surface
  ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
  ddsd.dwHeight = Height;
  ddsd.dwWidth = Width;
  DDError(pdd->CreateSurface(&ddsd, &pdds, NULL));
  return pdds;
}
//---------------------------------------------------------------------------
bool __fastcall DDDestroySurface(LPDIRECTDRAWSURFACE lpdds)
{
  if (lpdds != NULL) {
    lpdds->Release();
    lpdds = NULL;
    return true;
  }
  else return false;
}
//---------------------------------------------------------------------------
IDirectDrawSurface *DDLoadBitmap(IDirectDraw *pdd, char *F_Name, char *P_Name)
{
  const FPoint fpNULL = {0, 0};
  HBITMAP hbm;
  BITMAP bm;
  IDirectDrawSurface *pdds;

  hbm = LoadImage(GetModuleHandle(F_Name), P_Name, IMAGE_BITMAP,
   0, 0, LR_CREATEDIBSECTION);
                                                 //Image failed to load from
  if (hbm == NULL)                               //resource, load from file
   hbm = LoadImage(NULL, P_Name, IMAGE_BITMAP, 0, 0,
    LR_LOADFROMFILE | LR_CREATEDIBSECTION);

  if (hbm == NULL) {                             //Error Loading Image
    ShowMessage("Can't load image: " + (AnsiString) P_Name);
    exit(EXIT_FAILURE);                          //exit the program
    return NULL;
  }

  GetObject(hbm, sizeof(bm), &bm);               //get size of bitmap
  pdds = DDCreateSurface(pdd, bm.bmHeight, bm.bmWidth);
  DDCopyBitmap(pdds, hbm, fpNULL, fpNULL);       //Copy Bitmap
  DeleteObject(hbm);                             //Delete Bitmap
  return pdds;
} //DDLoadBitmap
//---------------------------------------------------------------------------
HRESULT DDReLoadBitmap(IDirectDrawSurface *pdds, char *F_Name)
{
  const FPoint fpNULL = {0, 0};
  HBITMAP hbm;
  HRESULT hr;

  hbm = LoadImage(GetModuleHandle(NULL), F_Name, IMAGE_BITMAP, 0, 0,
   LR_CREATEDIBSECTION);

  if (hbm == NULL) {                             //Error Loading Image
    ShowMessage("Can't reload bitmap");
    exit(EXIT_FAILURE);                          //exit the program
    return E_FAIL;
  }

  hr = DDCopyBitmap(pdds, hbm, fpNULL, fpNULL);
  DDError(hr);

  DeleteObject(hbm);
  return hr;
} //DDReLoadBitmap
//---------------------------------------------------------------------------
#pragma warn -pia-                             //Turn off PIA
extern "C" IDirectDrawPalette *DDLoadPalette(IDirectDraw *pdd, char *F_Name)
{
  IDirectDrawPalette *ddpal;
  int i;
  int n;
  int fh;
  HRSRC h;
  LPBITMAPINFOHEADER lpbi;
  PALETTEENTRY ape[256];
  RGBQUAD *prgb;

  for (i=0; i<256; i++) {
    ape[i].peRed   = (BYTE)(((i >> 5) & 0x07) * 255 / 7);
    ape[i].peGreen = (BYTE)(((i >> 2) & 0x07) * 255 / 7);
    ape[i].peBlue  = (BYTE)(((i >> 0) & 0x03) * 255 / 3);
    ape[i].peFlags = (BYTE)0;
  }
  if (F_Name && (h = FindResource(NULL, F_Name, RT_BITMAP))) {
    lpbi = (LPBITMAPINFOHEADER)LockResource(LoadResource(NULL, h));
    if (!lpbi) {
      ShowMessage("Can't Lock resource!;");
      exit(EXIT_FAILURE);                        //exit the program
    }
    prgb = (RGBQUAD*)((BYTE*)lpbi + lpbi->biSize);

    if (lpbi == NULL || lpbi->biSize < sizeof(BITMAPINFOHEADER)) n = 0;
    else if (lpbi->biBitCount > 8) n = 0;
    else if (lpbi->biClrUsed == 0) n = 1 << lpbi->biBitCount;
    else n = lpbi->biClrUsed;

    for(i = 0; i < n; i++ ) {
      ape[i].peRed   = prgb[i].rgbRed;
      ape[i].peGreen = prgb[i].rgbGreen;
      ape[i].peBlue  = prgb[i].rgbBlue;
      ape[i].peFlags = 0;
    }
  }
  else if (F_Name && (fh = _lopen(F_Name, OF_READ)) != -1) {
    BITMAPFILEHEADER bf;
    BITMAPINFOHEADER bi;

    _lread(fh, &bf, sizeof(bf));
    _lread(fh, &bi, sizeof(bi));
    _lread(fh, ape, sizeof(ape));
    _lclose(fh);

    if (bi.biSize != sizeof(BITMAPINFOHEADER)) n = 0;
    else if (bi.biBitCount > 8) n = 0;
    else if (bi.biClrUsed == 0) n = 1 << bi.biBitCount;
    else n = bi.biClrUsed;

    for(i = 0; i<n; i++ ) {
      BYTE r = ape[i].peRed;
      ape[i].peRed  = ape[i].peBlue;
      ape[i].peBlue = r;
    }
  }

  pdd->CreatePalette(DDPCAPS_8BIT, ape, &ddpal, NULL);

  return ddpal;
} //DDLoadPalette
#pragma warn -pia                                //Turn on PIA
//---------------------------------------------------------------------------
DWORD DDColorMatch(IDirectDrawSurface *pdds, COLORREF rgb)
{
  COLORREF rgbT;
  HDC hdc;
  DWORD dw = CLR_INVALID;
  DDSURFACEDESC ddsd;
  HRESULT hres;

  if (rgb != CLR_INVALID && pdds->GetDC(&hdc) == DD_OK) {
    rgbT = GetPixel(hdc, 0, 0);                  //save current pixel value
    SetPixel(hdc, 0, 0, rgb);                    //set the value
    pdds->ReleaseDC(hdc);
  }

  ddsd.dwSize = sizeof(ddsd);
  while ((hres = pdds->Lock(NULL, &ddsd, 0, NULL)) == DDERR_WASSTILLDRAWING);
  if (hres == DD_OK) {
    dw  = *(DWORD *)ddsd.lpSurface;
    if(ddsd.ddpfPixelFormat.dwRGBBitCount < 32)
    dw &= (1 << ddsd.ddpfPixelFormat.dwRGBBitCount)-1;
    pdds->Unlock(NULL);
  }
  if (rgb != CLR_INVALID && pdds->GetDC(&hdc) == DD_OK) {
    SetPixel(hdc, 0, 0, rgbT);
    pdds->ReleaseDC(hdc);
  }
  return dw;
} //DDColorMatch
//---------------------------------------------------------------------------
HRESULT DDSetColorKey(IDirectDrawSurface *pdds, COLORREF rgb)
{
  DDCOLORKEY ddck;
  ddck.dwColorSpaceLowValue  = DDColorMatch(pdds, rgb);
  ddck.dwColorSpaceHighValue = ddck.dwColorSpaceLowValue;
  return pdds->SetColorKey(DDCKEY_SRCBLT, &ddck);
} //DDSetColorKey
//---------------------------------------------------------------------------
FPoint DDGetImageSize(char *F_Name, char *P_Name)
{
  HBITMAP hbm;
  BITMAP bm;
  FPoint Size;

  hbm = LoadImage(GetModuleHandle(F_Name), P_Name, IMAGE_BITMAP,
   0, 0, LR_CREATEDIBSECTION);
                                                 //Image failed to load from
  if (hbm == NULL)                               //resource, load from file
   hbm = LoadImage(NULL, P_Name, IMAGE_BITMAP, 0, 0,
    LR_LOADFROMFILE | LR_CREATEDIBSECTION);

  GetObject(hbm, sizeof(bm), &bm);
  Size.X = bm.bmWidth;
  Size.Y = bm.bmHeight;
  return Size;
} //GetImageSize
//---------------------------------------------------------------------------
void __fastcall DDError(HRESULT d)
{
  char *S;
  if (d == DD_OK) return;
  if (d == DDERR_ALREADYINITIALIZED) S = "Object already initialised";
  else if (d == DDERR_CANNOTATTACHSURFACE) S = "Can't attach surface";
  else if (d == DDERR_CANNOTDETACHSURFACE) S = "Can't detach surface";
  else if (d == DDERR_CURRENTLYNOTAVAIL) S = "Support is not available";
  else if (d == DDERR_EXCEPTION) S = "Exception was encountered";
  else if (d == DDERR_GENERIC) S = "Generic failure";
  else if (d == DDERR_HEIGHTALIGN) S = "Rectangle out of allignment";
  else if (d == DDERR_INCOMPATIBLEPRIMARY) S = "Incompatible primary surface";
  else if (d == DDERR_INVALIDCAPS) S = "Caps bits are incorrect";
  else if (d == DDERR_INVALIDCLIPLIST) S = "DDraw does not support Cliplist";
  else if (d == DDERR_INVALIDMODE) S = "Can't set mode to 640x480";
  else if (d == DDERR_INVALIDOBJECT) S = "Invalid Object";
  else if (d == DDERR_INVALIDPARAMS) S = "Invalid Parameters";
  else if (d == DDERR_INVALIDPIXELFORMAT) S = "Invalid pixel format";
  else if (d == DDERR_INVALIDRECT) S = "Invalid Rectangle";
  else if (d == DDERR_LOCKEDSURFACES) S = "Sufrace is locked";
  else if (d == DDERR_NO3D) S = "No 3D";
  else if (d == DDERR_NOALPHAHW) S = "No Alpha acceleration";
  else if (d == DDERR_NOCLIPLIST) S = "No clip list available";
  else if (d == DDERR_NOCOLORCONVHW) S = "No Colour conversion";
  else if (d == DDERR_NOCOOPERATIVELEVELSET) S = "No cooperative level set";
  else if (d == DDERR_NOCOLORKEY) S = "No colour key available";
  else if (d == DDERR_NOCOLORKEYHW) S = "Hardware does not support colour key";
  else if (d == DDERR_NODIRECTDRAWSUPPORT) S = "No DDraw support";
  else if (d == DDERR_NOEXCLUSIVEMODE) S = "No exclusive mode";
  else if (d == DDERR_NOFLIPHW) S = "Can't flip surfaces";
  else if (d == DDERR_NOGDI) S = "No GDI present";
  else if (d == DDERR_NOMIRRORHW) S = "No hardware present";
  else if (d == DDERR_NOTFOUND) S = "Item was not found";
  else if (d == DDERR_NOOVERLAYHW) S = "No overlay hardware present";
  else if (d == DDERR_NORASTEROPHW) S = "No raster hardware present";
  else if (d == DDERR_NOROTATIONHW) S = "No rotation hardware present";
  else if (d == DDERR_NOSTRETCHHW) S = "No streching hardware present";
  else if (d == DDERR_NOT4BITCOLOR) S = "Not in 4Bit colour mode";
  else if (d == DDERR_NOT4BITCOLORINDEX) S = "Not in 4Bit colour index";
  else if (d == DDERR_NOT8BITCOLOR) S = "Not in 8Bit colour mode";
  else if (d == DDERR_NOTEXTUREHW) S = "No texture hardware present";
  else if (d == DDERR_NOVSYNCHW) S = "Hardware error";
  else if (d == DDERR_NOZBUFFERHW) S = "No hardware for Z buffer blitting";
  else if (d == DDERR_NOZOVERLAYHW) S = "No hardware for Z overlay";
  else if (d == DDERR_OUTOFCAPS) S = "Hardware has been allocated";
  else if (d == DDERR_OUTOFMEMORY) S = "Out of memory";
  else if (d == DDERR_OUTOFVIDEOMEMORY) S = "Out of Video memory";
  else if (d == DDERR_OVERLAYCANTCLIP) S = "No hardware for clipped overlays";
  else if (d == DDERR_OVERLAYCOLORKEYONLYONEACTIVE) S = "Can only have 1 colour key";
  else if (d == DDERR_PALETTEBUSY) S = "Pallete is busy";
  else if (d == DDERR_COLORKEYNOTSET) S = "Colour key not set";
  else if (d == DDERR_SURFACEALREADYATTACHED) S = "Surface already attached";
  else if (d == DDERR_SURFACEALREADYDEPENDENT) S = "Surface already dependent";
  else if (d == DDERR_SURFACEBUSY) S = "Surface is busy";
  else if (d == DDERR_CANTLOCKSURFACE) S = "Can't lock surface";
  else if (d == DDERR_SURFACEISOBSCURED) S = "Surface is obscured";
  else if (d == DDERR_SURFACELOST) S = "Surface has been lost";
  else if (d == DDERR_SURFACENOTATTACHED) S = "Surface is not attached";
  else if (d == DDERR_TOOBIGHEIGHT) S = "Height is too large";
  else if (d == DDERR_TOOBIGSIZE) S = "Size is too large";
  else if (d == DDERR_TOOBIGWIDTH) S = "Width is too large";
  else if (d == DDERR_UNSUPPORTED) S = "Action not supported";
  else if (d == DDERR_UNSUPPORTEDFORMAT) S = "Format not supported";
  else if (d == DDERR_UNSUPPORTEDMASK) S = "Unsupported mask";
  else if (d == DDERR_VERTICALBLANKINPROGRESS) S = "Vertical blank in progress";
  else if (d == DDERR_WASSTILLDRAWING) S = "Previous blit is incomplete";
  else if (d == DDERR_XALIGN) S = "Rectangle not horizontally aligned";
  else if (d == DDERR_INVALIDDIRECTDRAWGUID) S = "Invalid GUID";
  else if (d == DDERR_DIRECTDRAWALREADYCREATED) S = "Object has already been created";
  else if (d == DDERR_NODIRECTDRAWHW) S = "No DDraw hardware";
  else if (d == DDERR_PRIMARYSURFACEALREADYEXISTS)
   S = "There is already a primary surafce";
  else if (d == DDERR_NOEMULATION) S = "Software emulation not available";
  else if (d == DDERR_REGIONTOOSMALL) S = "Region is too small";
  else if (d == DDERR_CLIPPERISUSINGHWND) S = "Can't set clip list";
  else if (d == DDERR_NOCLIPPERATTACHED) S = "No clipper attached";
  else if (d == DDERR_NOHWND) S = "No HWND";
  else if (d == DDERR_HWNDSUBCLASSED) S = "DDraw can't restore";
  else if (d == DDERR_HWNDALREADYSET) S = "Can't reset HWND";
  else if (d == DDERR_NOPALETTEATTACHED) S = "No palette is attached";
  else if (d == DDERR_NOPALETTEHW) S = "No support for 4 or 8 Bit palettes";
  else if (d == DDERR_BLTFASTCANTCLIP) S = "Blit Fast can't clip";
  else if (d == DDERR_NOBLTHW) S = "No blitting available";
  else if (d == DDERR_NODDROPSHW) S = "No ROP hardware";
  else if (d == DDERR_OVERLAYNOTVISIBLE) S = "Overlay not visible";
  else if (d == DDERR_NOOVERLAYDEST) S = "No overlay destination";
  else if (d == DDERR_INVALIDPOSITION) S = "Invalid destination";
  else if (d == DDERR_NOTAOVERLAYSURFACE) S = "Not an overlay surface";
  else if (d == DDERR_EXCLUSIVEMODEALREADYSET) S = "Exclusive mode already set";
  else if (d == DDERR_NOTFLIPPABLE) S = "Surface not flippable";
  else if (d == DDERR_CANTDUPLICATE) S = "Can't duplicate surface";
  else if (d == DDERR_NOTLOCKED) S = "Surface was not locked";
  else if (d == DDERR_CANTCREATEDC) S = "Can't create any more DC's";
  else if (d == DDERR_NODC) S = "No DC was created";
  else if (d == DDERR_WRONGMODE) S = "Can't restore surface with a different mode";
  else if (d == DDERR_IMPLICITLYCREATED) S = "Can't restore surface";
  else if (d == DDERR_NOTPALETTIZED) S = "Surface is not palettized";
  else if (d == DDERR_UNSUPPORTEDMODE) S = "Display mode not supported";
  else if (d == DDERR_NOMIPMAPHW) S = "No mip-map hardware";
  else if (d == DDERR_INVALIDSURFACETYPE) S = "Invalid surface type";
  else if (d == DDERR_DCALREADYCREATED) S = "Can't return DC";
  else if (d == DDERR_CANTPAGELOCK) S = "Can't lock surface";
  else if (d == DDERR_CANTPAGEUNLOCK) S = "Can't unlock surface";
  else if (d == DDERR_NOTPAGELOCKED) S = "Surface was not locked";

  ShowMessage(S);                                //Show Error Message
  exit(EXIT_FAILURE);                            //exit the program
}
//---------------------------------------------------------------------------
};                                               //End namespace
//---------------------------------------------------------------------------

