//---------------------------------------------------------------------------
#ifndef E7_DDrawH
#define E7_DDrawH
//---------------------------------------------------------------------------
#include <ddraw.h>
#include <dsgnintf.hpp>
#include <Graphics.hpp>
#include <Forms.hpp>
#include <Controls.hpp>
#include <Classes.hpp>
#include <Windows.hpp>
#include <System.hpp>
//---------------------------------------------------------------------------
#ifndef __IPOINT
#define __IPOINT
typedef struct {int X, Y;} IPoint;               //Integer equivelent of TPoint
#endif

#ifndef __AGSRES
#define __AGSRES
enum TResolution {trNormal, tr320x240, tr400x300, tr512x384, tr640x480,
 tr800x600, tr960x720, tr1024x768, tr1152x864, tr1280x1024, tr1600x1200};
#endif
//Classes--------------------------------------------------------------------
class TE7_DDraw;
class TE7_DDSurface;
class TE7_Sprite;
//---------------------------------------------------------------------------
typedef void __fastcall (__closure *TDrawSceneEvent)(TObject* Sender,
 LPDIRECTDRAWSURFACE7 Surface);
//---------------------------------------------------------------------------
class PACKAGE TE7_Sprite
{
  private:
    int FPosition;
    bool FSingle;
  public:
    TList *Images;
    __property int Position = {read=FPosition};
    __fastcall TE7_Sprite(void);
    __fastcall ~TE7_Sprite(void);

    void __fastcall AddImage(HANDLE Dll, AnsiString FileName);
    void __fastcall AddImage(HANDLE Dll, AnsiString FileName,
     COLORREF Transparent);
    void __fastcall BlitFast(LPDIRECTDRAWSURFACE7 Dest,
     IPoint SourcePos, RECT SourceSize, LPRECT DestSize);
    void __fastcall BlitFx(LPDIRECTDRAWSURFACE7 Dest,
     IPoint SourcePos, RECT Size, DWORD Blt_Flags, LPDDBLTFX ddfx);
};
//---------------------------------------------------------------------------
class PACKAGE TE7_DDSurface : public TComponent
{
  friend TE7_DDraw;
private:
  TColor FBackColor;
  AnsiString FBackgroundBitmap;
  RECT FBackRect;
  bool FBlankScene;
  HANDLE FDLLFile;
  AnsiString FDLLName;
  TDrawSceneEvent FOnDrawScene;
  TNotifyEvent FOnSetupSurfaces;
  LPDIRECTDRAWPALETTE FPalette;
  bool FShowBitmap;
  LPDIRECTDRAWSURFACE7 FDDSurface;
  TColor FTransparentColor;

  LPDIRECTDRAWSURFACE7 CreateBitmap(HANDLE Bitmap);
  bool __fastcall CreateSurface(LPDIRECTDRAWSURFACE7 &pdds,
   AnsiString BitmapName, bool UsePalette);
  HANDLE __fastcall Get_Dib(HANDLE Instance, AnsiString S);
  AnsiString __fastcall GetFileName();

public:
  __property LPDIRECTDRAWSURFACE7 DDSurface={read=FDDSurface, write=FDDSurface};

  __fastcall TE7_DDSurface(TComponent* Owner);
  __fastcall ~TE7_DDSurface(void);
  void __fastcall DestroyObjects();
  void __fastcall DrawScene();
  HRESULT RestoreSurfaces();
  void __fastcall SetupSurfaces(TObject* Sender);
  void __fastcall WriteXY(int X, int Y, AnsiString S,
   COLORREF TextColor, COLORREF BackColor);

__published:
  __property TColor BackColor = {read=FBackColor, write=FBackColor};
  __property AnsiString BackgroundBitmap = {read=FBackgroundBitmap, write=FBackgroundBitmap};
  __property bool BlankScene = {read=FBlankScene, write=FBlankScene, nodefault};
  __property AnsiString DLLName = {read=FDLLName, write=FDLLName};
  __property bool ShowBitmap = {read=FShowBitmap, write=FShowBitmap, default=1};
  __property TDrawSceneEvent OnDrawScene = {read=FOnDrawScene, write=FOnDrawScene};
  __property Classes::TNotifyEvent OnSetupSurfaces = {read=FOnSetupSurfaces, write=FOnSetupSurfaces};
  __property TColor TransparentColor = {read=FTransparentColor, write=FTransparentColor};
};
//---------------------------------------------------------------------------
class PACKAGE TE7_DDraw : public TComponent
{
  private:
    bool FActive;
    LPDIRECTDRAWSURFACE7 FBackSurface;           //Back Surface
    int FBitsPerPixel;
    LPDIRECTDRAWCLIPPER FClipper;                //Clipper
    LPDIRECTDRAW7 FDirectDraw;                   //DD
    bool FExclusive;
    bool FFirstTime;
    HWND FHandle;
    LPDIRECTDRAWSURFACE7 FPrimarySurface;        //Primary Surface
    TResolution FResolution;
    bool FTimerOdd;
    int FTimerInterval;
    bool FUseTimer;
    TNotifyEvent FPaintProc;
    TE7_DDSurface *FE_DDSurface;                  //DDSurface

    void CreateBack();
    void CreateDD();
    void CreatePrimary();
    bool __fastcall MakeItSo(HRESULT DDResult);
    void __fastcall SetScene(TE7_DDSurface* E_DDSurface);
    void SetScreenSize();

  protected:
    int ChoseColourDepth(int Depth);
    void __fastcall DrawBitmaps(void);
    HRESULT RestoreSurfaces();
    virtual void __fastcall Notification(TComponent* AComponent, TOperation Operation);

  public:
    RECT ScreenRect;

    __fastcall TE7_DDraw(Classes::TComponent* Owner);
    __fastcall ~TE7_DDraw(void);
    void __fastcall DDTest(HRESULT ddrval, AnsiString Orign);
    void __fastcall EndExclusive();
    void __fastcall ErrorEvent(AnsiString Msg);
    void __fastcall Flip();
    AnsiString GetVersion() {return "Engine7: Version 7.2";};
    void __fastcall Init();

    __property bool Active = {read=FActive, write=FActive};
    __property LPDIRECTDRAW7 ADirectDraw={read=FDirectDraw, write=FDirectDraw};
    __property LPDIRECTDRAWSURFACE7 BackSurface = {read=FBackSurface, write=FBackSurface};
    __property LPDIRECTDRAWSURFACE7 PrimarySurface = {read=FPrimarySurface, write=FPrimarySurface};
    __property TNotifyEvent OnDrawBitmap = {read=FPaintProc, write=FPaintProc};

  __published:
    __property int BitsPerPixel = {read=FBitsPerPixel, write=FBitsPerPixel};
    __property TE7_DDSurface* E_DDSurface = {read=FE_DDSurface, write=SetScene};
    __property bool Exclusive = {read=FExclusive, write=FExclusive};
    __property int TimerInterval = {read=FTimerInterval, write=FTimerInterval, default=1000};
    __property bool UseTimer = {read=FUseTimer, write=FUseTimer};
    __property TResolution Resolution={read=FResolution, write=FResolution};
};
//---------------------------------------------------------------------------
namespace DDUtils {
  void __fastcall DDBlitFast(LPDIRECTDRAWSURFACE7 Dest, LPDIRECTDRAWSURFACE7 Source,
   IPoint SourcePos, RECT SourceSize, LPRECT DestSize);
  void __fastcall DDBlitFx(LPDIRECTDRAWSURFACE7 Dest, LPDIRECTDRAWSURFACE7 Source,
   IPoint SourcePos, RECT Size, DWORD Blt_Flags, LPDDBLTFX ddfx);
  HRESULT DDCopyBitmap(LPDIRECTDRAWSURFACE7 lpdds, HBITMAP hbm, RECT Grid);
  LPDIRECTDRAWSURFACE7 DDCreateSurface(DWORD Width, DWORD Height);
  DWORD DDColorMatch(LPDIRECTDRAWSURFACE7 lpdds, COLORREF rgb);
  void __fastcall DDDestroySurface(LPDIRECTDRAWSURFACE7 &lpdds);
  IPoint DDGetImageSize(HANDLE Dll, AnsiString Pic);
  LPDIRECTDRAWSURFACE7 DDLoadBitmap(HANDLE Dll, AnsiString Pic, bool Transparent,
   COLORREF TransColour);
  LPDIRECTDRAWPALETTE DDLoadPalette(HANDLE Instance,
   LPDIRECTDRAW7 pdd, const AnsiString BitmapName);
  void __fastcall DDReloadBitmap(LPDIRECTDRAWSURFACE7 pdds, HANDLE Dll,
   AnsiString Pic);
  HRESULT DDSetColorKey(LPDIRECTDRAWSURFACE7 lpdds, COLORREF Colour);
};
//---------------------------------------------------------------------------
using namespace DDUtils;
#define Timer1 (Byte)(1)
//end unit-------------------------------------------------------------------
#endif
