//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "E7_Sound.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
static inline void ValidCtrCheck(TE7_Sound *)
{
  new TE7_Sound(NULL);
}
//---------------------------------------------------------------------------
__fastcall TE7_Sound::TE7_Sound(TComponent *Owner)
:TComponent(Owner)
{
  FHandle = ((TWinControl *)(Owner))->Handle;
}
//---------------------------------------------------------------------------
__fastcall TE7_Sound::~TE7_Sound()
{
  Close();
  FreeDirectSound();
}
//---------------------------------------------------------------------------
void TE7_Sound::CreateStaticBuffer(AnsiString F_Name)
{
  HRESULT dsrval;
  DSBUFFERDESC dsbd;

  if(Open(F_Name) != DS_OK) {
    ShowMessage("Bad wave file.");
    return;
  }

  ZeroMemory(&dsbd, sizeof(dsbd));
  dsbd.dwSize = sizeof(dsbd);
  dsbd.dwFlags = DSBCAPS_STATIC;
  dsbd.dwBufferBytes = FckIn.cksize;
  dsbd.lpwfxFormat = Fpwfx;

  dsrval = FDirectSound->CreateSoundBuffer( &dsbd, &FDSBuffer, NULL);
  if (dsrval != DS_OK) {
    ShowMessage("Can't load buffer");
    return;
  }

  FBufferBytes = dsbd.dwBufferBytes;             //Take Size
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::FillBuffer()
{
  HRESULT dsrval;
  BYTE* lpbWavData;                              //Actual wav data
  UINT cbWavSize;                                //Size of data
  void* lpvData  = NULL;
  void* lpvData2 = NULL;
  DWORD dwLength;
  DWORD dwLength2;
  int WaveFileSize;

  WaveFileSize = FckIn.cksize;

  lpbWavData = new BYTE[WaveFileSize];           //Allocate Buffer
  if(lpbWavData == NULL) return E_OUTOFMEMORY;

  dsrval = Read(WaveFileSize, lpbWavData, &cbWavSize);
  if (dsrval != DS_OK) return dsrval;

  Reset();                                       //Reset file to beginning

  dsrval = FDSBuffer->Lock(0, FBufferBytes, &lpvData, &dwLength, &lpvData2,
   &dwLength2, 0L);
  if (dsrval != DS_OK) return dsrval;

  memcpy(lpvData, lpbWavData, FBufferBytes);

  FDSBuffer->Unlock(lpvData, FBufferBytes, NULL, 0);
  lpvData = NULL;

  delete lpbWavData;

  return DS_OK;
}
//---------------------------------------------------------------------------
void __fastcall TE7_Sound::FreeDirectSound()
{
  if (GetStatus() & DSBSTATUS_PLAYING) StopBuffer();

  if (FDSBuffer) {
    FDSBuffer->Release();
    FDSBuffer = NULL;
  }
  if (FDirectSound) {
    FDirectSound->Release();
    FDirectSound = NULL;
  }

  CoUninitialize();                              //Release COM
}
//---------------------------------------------------------------------------
DWORD TE7_Sound::GetStatus()
{
  DWORD dwStatus = 0;

  if(FDSBuffer) FDSBuffer->GetStatus(&dwStatus);
  else dwStatus = DSBSTATUS_BUFFERLOST;

  return dwStatus;
}
//---------------------------------------------------------------------------
void __fastcall TE7_Sound::Init()
{
  HRESULT dsrval;

  dsrval = CoInitialize(NULL);
  if (dsrval != DS_OK) throw Exception("Can't initialize COM");

  dsrval = DirectSoundCreate(NULL, &FDirectSound, NULL);
  if (dsrval != DS_OK) throw Exception("Can't create Direct Sound");

  dsrval = FDirectSound->SetCooperativeLevel(FHandle, DSSCL_PRIORITY);
  if (dsrval != DS_OK) throw Exception("Can't set coopertive level");

}
//---------------------------------------------------------------------------
void __fastcall TE7_Sound::LoadWaveFile(AnsiString F_Name)
{
  if (FFileName == F_Name) return;               //Don't reopen file
  else FFileName = F_Name;

  CreateStaticBuffer(F_Name);
  FillBuffer();
}
//---------------------------------------------------------------------------
void __fastcall TE7_Sound::PlayBuffer(bool Looped)
{
  HRESULT dsrval;
  DWORD dwFlags;
  DWORD DBStatus;

  DBStatus = GetStatus();
  if (DBStatus & DSBSTATUS_BUFFERLOST) {         //Lost the buffer
    dsrval = RestoreBuffer();                    //Restore Buffer
    if (dsrval != DS_OK) {
      ShowMessage("Can't restore buffer");
      return;
    }
  }
  else if (DBStatus & DSBSTATUS_PLAYING) StopBuffer();
  else if (DBStatus & DSBSTATUS_LOOPING) StopBuffer();

  if (Looped) dwFlags = DSBPLAY_LOOPING;         //Set flags
  else dwFlags = 0;

  dsrval = FDSBuffer->Play(0, 0, dwFlags);       //Play the buffer
  if (dsrval != DS_OK) {
    ShowMessage("Can't play buffer");
    return;
  }
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::RestoreBuffer()
{
  HRESULT dsrval;
  DWORD dwStatus;
  int Tries = 1;

  if(FDSBuffer == NULL) return DS_OK;

  dwStatus = GetStatus();

  if(dwStatus & DSBSTATUS_BUFFERLOST) {
    dsrval = FDSBuffer->Restore();
    while (Tries < 10) {                         //Attempt to retrieve buffer
      if(dsrval == DSERR_BUFFERLOST) Sleep(20);
      dsrval = FDSBuffer->Restore();
    }
    if (dsrval != DS_OK) return dsrval;

    dsrval = FillBuffer();
    if (dsrval != DS_OK) return dsrval;
  }

  return DS_OK;
}
//---------------------------------------------------------------------------
void __fastcall TE7_Sound::StopBuffer()
{
  if(FDSBuffer == NULL) return;
  
  FDSBuffer->Stop();
  FDSBuffer->SetCurrentPosition(0L);
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::ReadMMIO(HMMIO hmmioIn, MMCKINFO* pckInRIFF, WAVEFORMATEX** ppwfxInfo)
{
  MMCKINFO ckIn;                                 //chunk info for general use
  PCMWAVEFORMAT pcmWaveFormat;                   //Temp PCM structure to load in

  *ppwfxInfo = NULL;

  if(mmioDescend(hmmioIn, pckInRIFF, NULL, 0) != 0) return E_FAIL;

  if((pckInRIFF->ckid != FOURCC_RIFF) ||
     (pckInRIFF->fccType != mmioFOURCC('W', 'A', 'V', 'E'))) return E_FAIL;

  //Search the input file for for the 'fmt' chunk
  ckIn.ckid = mmioFOURCC('f', 'm', 't', ' ');
  if(mmioDescend(hmmioIn, &ckIn, pckInRIFF, MMIO_FINDCHUNK) != 0) return E_FAIL;

  //Expect the 'fmt' chunk to be at least as large as <PCMWAVEFORMAT>
  //if there are extra parameters at the end ignore them
  if(ckIn.cksize < (long) sizeof(PCMWAVEFORMAT)) return E_FAIL;

  // Read the 'fmt' chunk into <pcmWaveFormat>
  if(mmioRead(hmmioIn, (HPSTR) &pcmWaveFormat,
   sizeof(pcmWaveFormat)) != (long)sizeof(pcmWaveFormat)) return E_FAIL;

  //Allocate the waveformatex, but if its not pcm format, read the next
  //word, and thats how many extra bytes to allocate.
  if(pcmWaveFormat.wf.wFormatTag == WAVE_FORMAT_PCM) {
    if((*ppwfxInfo = new WAVEFORMATEX) == NULL) return E_FAIL;

    //Copy the bytes from the pcm structure to the waveformatex structure
    memcpy(*ppwfxInfo, &pcmWaveFormat, sizeof(pcmWaveFormat));
    (*ppwfxInfo)->cbSize = 0;
  }
  else {
    // Read in length of extra bytes.
    WORD cbExtraBytes = 0L;
    if(mmioRead(hmmioIn, (char*)&cbExtraBytes, sizeof(WORD)) != (long)sizeof(WORD))
     return E_FAIL;

    *ppwfxInfo = (WAVEFORMATEX*)new char[sizeof(WAVEFORMATEX) + cbExtraBytes];
    if(*ppwfxInfo == NULL) return E_FAIL;

    // Copy the bytes from the pcm structure to the waveformatex structure
    memcpy(*ppwfxInfo, &pcmWaveFormat, sizeof(pcmWaveFormat));
    (*ppwfxInfo)->cbSize = cbExtraBytes;

    // Now, read those extra bytes into the structure, if cbExtraAlloc != 0.
     if(mmioRead(hmmioIn, (char*)(((BYTE*)&((*ppwfxInfo)->cbSize))+sizeof(WORD)),
      cbExtraBytes ) != cbExtraBytes) {
       delete *ppwfxInfo;
       *ppwfxInfo = NULL;
       return E_FAIL;
     }
   }

   //Ascend the input file out of the 'fmt ' chunk.
   if(mmioAscend( hmmioIn, &ckIn, 0) != 0) {
     delete *ppwfxInfo;
     *ppwfxInfo = NULL;
     return E_FAIL;
  }

  return S_OK;
}
//-----------------------------------------------------------------------------
HRESULT TE7_Sound::WaveOpenFile(char* FileName, HMMIO* phmmioIn,
 WAVEFORMATEX** ppwfxInfo, MMCKINFO* pckInRIFF )
{
  HRESULT dsrval;
  HMMIO hmmioIn = NULL;

  if((hmmioIn = mmioOpen(FileName, NULL, MMIO_ALLOCBUF|MMIO_READ)) == NULL)
    return E_FAIL;

  dsrval = ReadMMIO(hmmioIn, pckInRIFF, ppwfxInfo);
  if (dsrval != 0) {
    mmioClose(hmmioIn, 0);
    return dsrval;
  }

  *phmmioIn = hmmioIn;

  return S_OK;
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::WaveStartDataRead( HMMIO* phmmioIn, MMCKINFO* pckIn,
 MMCKINFO* pckInRIFF)
{
  if(mmioSeek(*phmmioIn, pckInRIFF->dwDataOffset + sizeof(FOURCC), SEEK_SET ) == -1)
   return E_FAIL;

  //Search the input file for for the 'data' chunk.
  pckIn->ckid = mmioFOURCC('d', 'a', 't', 'a');
  if(mmioDescend(*phmmioIn, pckIn, pckInRIFF, MMIO_FINDCHUNK) != 0)
   return E_FAIL;

  return S_OK;
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::WaveReadFile( HMMIO hmmioIn, UINT cbRead, BYTE* pbDest,
 MMCKINFO* pckIn, UINT* cbActualRead )
{
  MMIOINFO mmioinfoIn;                           //current status of <hmmioIn>
  UINT cbDataIn;

  *cbActualRead = 0;

  if(mmioGetInfo(hmmioIn, &mmioinfoIn, 0) != 0) return E_FAIL;

  cbDataIn = cbRead;
  if(cbDataIn > pckIn->cksize) cbDataIn = pckIn->cksize;
  pckIn->cksize -= cbDataIn;

  for(DWORD cT = 0; cT < cbDataIn; cT++) {
    // Copy the bytes from the io to the buffer.
    if(mmioinfoIn.pchNext == mmioinfoIn.pchEndRead) {
      if(mmioAdvance( hmmioIn, &mmioinfoIn, MMIO_READ) != 0) return E_FAIL;
      if(mmioinfoIn.pchNext == mmioinfoIn.pchEndRead) return E_FAIL;
    }
    // Actual copy.
    *((BYTE*)pbDest+cT) = *((BYTE*)mmioinfoIn.pchNext);
    mmioinfoIn.pchNext++;
  }

  if(mmioSetInfo(hmmioIn, &mmioinfoIn, 0) != 0) return E_FAIL;

  *cbActualRead = cbDataIn;

  return S_OK;
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::Open(AnsiString FName)
{
  HRESULT dsrval;

  delete Fpwfx;

  dsrval = WaveOpenFile(FName.c_str(), &FhmmioIn, &Fpwfx, &FckInRiff);
  if (FAILED(dsrval)) return dsrval;

  dsrval = Reset();
  if(FAILED(dsrval)) return dsrval;

  return S_OK;
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::Reset()
{
  return WaveStartDataRead(&FhmmioIn, &FckIn, &FckInRiff);
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::Read(UINT nSizeToRead, Byte* pbData, UINT* pnSizeRead )
{
  return WaveReadFile(FhmmioIn, nSizeToRead, pbData, &FckIn, pnSizeRead );
}
//---------------------------------------------------------------------------
HRESULT TE7_Sound::Close()
{
  return mmioClose(FhmmioIn, 0);
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
namespace E7_sound
{
    void __fastcall PACKAGE Register()
    {
        TComponentClass classes[1] = {__classid(TE7_Sound)};
        RegisterComponents("Engine7", classes, 0);
    }
}
//---------------------------------------------------------------------------
